from .issue import states
from .utils import str2date
from .repo_stuff import Repo


class Event(object):
    def __init__(self, ev, label2val: dict = {}):
        self.date = str2date(ev['created_at'])
        self.id = ev.get('id', None)
        self.event = ev['event']
        self.label = ev.get('label', {}).get('name', None)
        self.label_val = label2val.get(self.label, None)
        self.assignee = ev.get('assignee', {}).get('login', None)

        src = ev.get('source', {})
        issue = src.get('issue', {})

        self.source_title = None
        self.source_number = None
        self.source_closed_at = None
        self.source_state = None
        self.source_is_pr = None
        self.source_repo = None
        if src:
            owner, repo = issue['repository_url'].split('/')[-2:]
            self.source_repo = Repo(owner, repo)
            self.source_title = issue.get('title', None)
            self.source_number = issue.get('number', None)
            self.source_closed_at = str2date(self.source_closed_at)
            self.source_state = states[issue.get('state', None)]
            self.source_is_pr = 'pull_request' in issue
            self.source_body = issue['body']



import concurrent.futures
import logging
import sys
from datetime import date, datetime, timedelta
from time import time
from typing import List, Tuple, TypeVar

from dateutil.relativedelta import WE, relativedelta
from rake_nltk import Rake
from tqdm import tqdm

from .board_card import Card
from .ghdriver import GithubHelper
from .issue import states

nl = '\n'


class Sprint(object):
    def __init__(self, access_token, repos, project_name, ignore_columns, keep_columns, login_name_mapper='', start=None, end=None, week_number=None, special_tags='', verbosity=0):
        self.ghh = GithubHelper(access_token, repos, project_name, ignore_columns, keep_columns, login_name_mapper=login_name_mapper, special_tags=special_tags, verbosity=verbosity)

        # logging
        self.logger = logging.Logger('sprint')
        self.logger.setLevel(logging.INFO)

        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = logging.Formatter(format)

        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        self.logger.info('starting')

        sprint_start_day = WE
        start_hour = 9
        end_hour = 19
        sprint_length = 7
        if not week_number:
            today = datetime.now()
            last_wed = today + relativedelta(weekday=sprint_start_day(-1))
            start = last_wed
        else:
            start = date(date.today().year, 1, 1)+timedelta(weeks=week_number) + relativedelta(weekday=sprint_start_day(-1))
            start = datetime.combine(start, datetime.min.time())

        start = start.replace(hour=start_hour, minute=0, microsecond=0, second=0)
        self.date_start = start
        if not end:
            self.date_end = self.date_start + timedelta(days=sprint_length-1)
        else:
            self.date_end = end
        self.date_end = self.date_end.replace(hour=end_hour, minute=0, microsecond=0, second=0)

        if self.date_end <= self.date_start:
            raise ValueError('end is set incorrectly')

    def set_pokered_leftovers(self, all_cards: List[Card]):
        self.all_pokered_leftover = [c for c in all_cards if c.has_label_val_created_within(
            self.date_start-timedelta(days=1), self.date_start)]

    def set_pokered(self, all_cards: List[Card]):
        self.all_pokered_cards = [c for c in all_cards if c.has_label_val_created_within(
            self.date_start, self.date_start+timedelta(hours=8))]

    def set_unchanged_stories(self, all_cards: List[Card]):
        self.all_stale_cards = [c for c in all_cards if c.col.name in self.ghh.cols_keep and c not in self.all_pokered_cards and c not in self.all_pokered_leftover]

    def set_repokered(self, all_cards: List[Card]):
        self.all_repokered_cards = [c for c in all_cards if c.has_label_val_created_within(
            self.date_end-timedelta(hours=12), self.date_end)]

    def set_PRs_without_stories(self, all_PRs):
        self.PRs_without_issues = [pr for pr in all_PRs if len(pr.closes) == 0 and self.date_start < pr.created_at < self.date_end]


    def fetch_all_data(self):
        start_0 = time()
        # to know that is in not yet in production
        # prod_sha = self.ghh.get_production_sha()
        # master_sha = self.ghh.get_master_sha()
        # compare = self.ghh.compare_commits(prod_sha, master_sha)
        # self.logger.info('time for production tag difference:\t{:.1f}s'.format(time() - start_0))

        # get project_id
        start = time()
        found_matching_name = False
        with tqdm(total=len(self.ghh.repos)) as pbar:
            for repo in self.ghh.repos:
                pbar.set_description(f'Checking {repo.owner}/{repo.name}')
                prjs = self.ghh.get_all_projects(repo)
                for project in prjs:
                    if self.ghh.project_name.lower() == project['name'].lower():
                        self.ghh.project_id = project['id']
                        found_matching_name = True
                        pbar.update(len(self.ghh.repos)-pbar.n)
                        break
                if found_matching_name:
                    break
                pbar.update()
        if not found_matching_name:
            raise RuntimeError(f'No suitable project called "{self.ghh.project_name}" found in {self.ghh.repos}')
        self.logger.info('time for getting project id:\t{:.1f}s'.format(time() - start))

        # get all columns in project
        start = time()
        cols = self.ghh.get_all_columns()
        self.logger.info('time for board columns:\t{:.1f}s'.format(time() - start))

        # get all cards from board and columns
        all_cards = []
        start = time()
        with tqdm(total=len(cols)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(max_workers=30) as executor:
                for col, cards in zip(cols, executor.map(self.ghh.get_all_cards, cols)):
                    pbar.set_description(f'Fetched cards for {col.name}')
                    all_cards.extend(cards)
                    pbar.update()

        self.logger.info('time for all board cards:\t{:.1f}s'.format(time() - start))

        # recognize related repos, add it to config
        self.ghh.repos = self.ghh.repos + list(set(c.repo for c in all_cards if c.repo != None and c.repo not in self.ghh.repos))

        # PRs
        start = time()
        pull_requests = self.ghh.get_all_PRs()
        self.logger.info('time for PRs:\t{:.1f}s'.format(time() - start))

        start = time()
        with tqdm(total=len(pull_requests)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(max_workers=30) as executor:
                for pr, revs in zip(pull_requests, executor.map(self.ghh.fetch_PR_reviews, pull_requests)):
                    pbar.set_description(f'Fetched PR reviews for #{pr.number}')
                    pr.add_reviews(revs)
                    pbar.update()
        self.logger.info('time for PR reviews:\t{:.1f}s'.format(time() - start))

        start = time()
        with tqdm(total=len(pull_requests)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(max_workers=30) as executor:
                for pr, data in zip(pull_requests, executor.map(self.ghh.fetch_PR_data, pull_requests)):
                    pbar.set_description(f'Fetched PR data for #{pr.number}')
                    pr.update(data)
                    pbar.update()
        self.logger.info('time for PR data:\t{:.1f}s'.format(time() - start))

        # get all events
        start = time()
        cards_with_issues = [c for c in all_cards if c.has_issue]
        with tqdm(total=len(cards_with_issues)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
                for card, events in zip(cards_with_issues, executor.map(self.ghh.get_all_events, [c.issue_num for c in cards_with_issues])):
                    pbar.set_description(f'Fetched events for #{card.id}')
                    card.set_events(events)
                    pbar.update()
        self.logger.info('time for all board card events:\t{:.1f}s'.format(time() - start))

        # issue data
        start = time()
        with tqdm(total=len(cards_with_issues)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(max_workers=30) as executor:
                for card, issue in zip(cards_with_issues, executor.map(self.ghh.get_single_issue_from_card, cards_with_issues)):
                    pbar.set_description(f'Fetched issue data for #{card.id}')
                    card.set_issue(issue)
                    pbar.update()
        self.logger.info('time for all board card issues:\t{:.1f}s'.format(time() - start))

        # total
        self.logger.info('total time:\t{:.1f}s'.format(time() - start_0))

        self.set_pokered_leftovers(cards_with_issues)
        self.set_pokered(cards_with_issues)
        self.set_repokered(cards_with_issues)
        self.set_unchanged_stories(cards_with_issues)
        self.prs = pull_requests
        self.set_PRs_without_stories(pull_requests)

    def get_all_stories(self):
        return self.all_pokered_cards + self.all_pokered_leftover + self.all_stale_cards + self.all_repokered_cards

    def print_report(self):
        def print_pr(pr, print_user):
            user = pr.user_name if print_user else ''
            special_tags = [tag for tag in pr.labels if tag['name'] in self.ghh.special_tags]
            special_tags_str = ''
            if special_tags:
                special_tags_str = ', '.join(f'[**{tag["name"]}**]' for tag in special_tags) + ', '
            return f' - PR {pr.get_state()} [{pr.repo.name} #{pr.number} {pr.title}]({pr.url} ) {user}\n' + \
                   f'      - {special_tags_str}+{pr.additions}, -{pr.deletions}, files changed: {pr.changed_files}\n' + \
                   f'      - Reviews: {pr.get_review_state(self.ghh.login_name_mapper)}'

        def print_story(story):
            txt = ''
            leftover = story.get_pokered_value(
                self.date_start-timedelta(days=1), self.date_start)
            poker = story.get_pokered_value(
                self.date_start, self.date_start+timedelta(hours=8)) or story.get_current_value()
            repoker = story.get_pokered_value(
                self.date_end-timedelta(hours=5), self.date_end) or '?'
            poker_repoker = '{}({})'.format(repoker, poker or leftover)
            special_tags = ' ' + ', '.join(f'[**{tag["name"]}**]' for tag in story.issue.labels if tag['name'] in self.ghh.special_tags)
            txt += '- {}{} [**{}**] [**{}**]({} ) {}'.format(
                story.get_state(),
                special_tags,
                poker_repoker,
                story.issue.title.strip(),
                story.issue.url,
                ', '.join(story.get_assignees(self.ghh.login_name_mapper))
            )
            pr_block = []
            for pr_event in [ev for ev in story.events if ev.source_is_pr]:
                merged_txt = ''
                prs = [pr for pr in self.prs if pr.number == pr_event.source_number and pr.repo == pr_event.source_repo]
                if len(prs) == 0:
                    pr = self.ghh.get_pr(pr_event.source_repo, pr_event.source_number)
                else:
                    pr = prs[0]
                pr_block.append('  ' + print_pr(pr, print_user=False))
            if len(pr_block) > 0:
                txt += nl
                txt += nl.join(pr_block)
            return txt

        # report list, will be joined at the end with newline character
        rprt = []

        # title by using word frequency
        all_titles = '. '.join([c.issue.title.strip() for c in self.get_all_stories()])
        r = Rake()
        r.extract_keywords_from_text(all_titles)
        word_degs = r.get_word_degrees()
        sorted_tuples = sorted(word_degs.items(), key=lambda item: item[1], reverse=True)
        rprt.append('# '+', '.join([t[0] for t in sorted_tuples[:10]]))
        rprt.append('')

        # date range: 06. February - 12. February 2019
        rprt.append('{} - {}'.format(self.date_start.strftime("%d. %B"),
                                     self.date_end.strftime("%d. %B %Y")))
        rprt.append('')

        # Velocity
        rprt.append('Velocity: **X**')
        rprt.append('')

        # Stories
        if self.all_pokered_leftover:
            rprt.append('# Leftover stories from last week')
            for story in self.all_pokered_leftover:
                rprt.append(print_story(story))
                rprt.append('')

        if self.all_pokered_cards:
            rprt.append('# Stories of the week')
            for story in self.all_pokered_cards:
                rprt.append(print_story(story))
                rprt.append('')

        if self.all_stale_cards:
            rprt.append('# Unchanged stories')
            for story in self.all_stale_cards:
                rprt.append(print_story(story))
                rprt.append('')

        # PRs without story:
        if self.PRs_without_issues:
            rprt.append('# PRs without issue')
            for pr in self.PRs_without_issues:
                rprt.append(print_pr(pr, print_user=True))

        return nl.join(rprt)




from .utils import str2date

class Repo(object):
    def __init__(self, owner:str, name:str):
        self.owner = owner
        self.name = name

    def __hash__(self):
        return hash(self.owner+self.name)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.owner == other.owner and self.name == other.name

    @classmethod
    def from_http(cls,url):
        if not url.startswith('http'):
            raise ValueError('{} does not start with `http`')
        owner, name = url.split('/')[-4:-2]
        return cls(owner, name)

class Commit(object):
    def __init__(self, commit):
        self.message = commit['commit']['message']
        self.sha = commit['commit']['tree']['sha']
        self.committer_name = commit['commit']['committer']['name']
        self.committer_date = str2date(commit['commit']['committer']['date'])

        pattern = ' \\(#(\\d+)\\)$'
        pr_num = None
        for pr_num in re.findall(pattern, self.message):
            self.pr_num = pr_num



from enum import Enum

from .repo_stuff import Repo
from .utils import str2date


class states(Enum):
    unknown = 0
    open = 1
    closed = 2


class Issue(object):
    def __init__(self, repo: Repo, issue_dict: dict):
        self.repo = repo
        self.closed_at = str2date(issue_dict.get('closed_at', None))
        self.id = issue_dict.get('id', None)
        self.number = issue_dict.get('number', None)
        self.title = issue_dict.get('title', None)
        self.url = issue_dict.get('html_url', None)
        self.assignees = issue_dict.get('assignees', [])
        self.labels = issue_dict.get('labels', [])
        state_string = issue_dict.get('state', None)
        self.state = None
        if state_string:
            self.state = states[state_string]

    @classmethod
    def from_closes_line(cls, line: str, default_repo: Repo):
        # https://github.com/fedger/menu-service/issues/857
        line = line.strip()
        if not line.lower().startswith('- closes '):
            raise ValueError('this line `{}` is not like `- closes [...]`'.format(line))

        http = 'http'
        if http in line:
            url = http+line.split(http)[-1]
            url = url.replace(')','')
            try:
                number = int(url.split('/')[-1])
            except:
                pass
                number = -1
            repo = Repo.from_http(url)
            return cls(repo, {'number': number, 'html_url': url})
        elif line.endswith('none'):
            return None
        elif '<' in line or '>' in line:
            return None
        else:
            if '#' not in line:
                raise ValueError('this line `{}` does not have a `#` in it'.format(line))

            number = int(line.split('#')[1])
            return cls(default_repo, {'number': number})



import concurrent.futures
import datetime
import os
import sys
import typing
from datetime import date, datetime, timedelta
from time import time
from typing import List, Tuple, TypeVar

import click

from ghsprint.sprint import Sprint

os.putenv('PYTHONIOENCODING', 'UTF-8')


@click.command()
@click.option('--ignore-columns', default='', help='list of columns names to ignore, separated by commas')
@click.option('-v', '--verbose', count=True)
@click.option('-w', '--week')
@click.option('--keep-columns', default='', help='list of columns names to always have in report, separated by commas')
@click.option('--login-name-mapper', default='', help='<login>:<name>,...')
@click.option('--special-tags', default='', help='<tag string>,...')
@click.argument('access-token')
@click.argument('project-name')
@click.argument('repos', nargs=-1)
def start_sprint(access_token, repos, project_name, ignore_columns, keep_columns, verbose, week, login_name_mapper, special_tags):
    """Fetches all data from github to create the sprint report."""
    if week:
        week = int(week)
    sp = Sprint(access_token, repos, project_name, ignore_columns, keep_columns, login_name_mapper=login_name_mapper, week_number=week, verbosity=verbose, special_tags=special_tags)
    sp.fetch_all_data()
    report = sp.print_report()
    print(report)


if __name__ == '__main__':
    start_sprint()



import concurrent.futures
import datetime
import typing
from datetime import date, datetime, timedelta
from time import time
from typing import Dict, List, Tuple, TypeVar

from .board_card import Card
from .board_stuff import Column
from .issue import Issue
from .issue_event import Event
from .pr_stuff import PR, Review
from .repo_stuff import Commit, Repo
from .request_session import requests_retry_session
from .utils import str2date


class GithubHelper(object):
    def __init__(self, access_token: str, repos: List[str], project_name: str, ignore_columns: str, keep_columns: str, login_name_mapper: str, special_tags: str, verbosity=0):
        self.access_token = access_token
        self.headers = {'Accept': 'application/vnd.github.inertia-preview+json'}

        if repos is None:
            raise ValueError('set `repos` in github-scope')

        if len(repos) == 0:
            raise ValueError('set at least one repo in `repos` in github-scope')

        self.repos = [Repo(x[0], x[1]) for x in (x.split("/") for x in repos) if x[1] != 5]

        self.project_name = project_name
        self.cols_ignore = ignore_columns.split(',')
        self.cols_keep = keep_columns.split(',')

        self.login_name_mapper = dict(entry.split(':') for entry in login_name_mapper.split(','))
        self.login_name_mapper = {k.lower(): v for k, v in self.login_name_mapper.items()}
        self.special_tags = special_tags.split(',')
        self.label2val = {'0': 0, '½': 0.5, '1': 1, '2': 2, '3': 3, '5': 5, '8': 8, '13': 13, '25': 25, '50': 50, '100': 100}

    def get_all_projects(self, repo: Repo):
        url = 'https://api.github.com/repos/{}/{}/projects?access_token={}'.format(repo.owner, repo.name, self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)
        if r.status_code == 200:
            return r.json()
        else:
            raise RuntimeError(f'Connection error fetching projects for {repo}')

    def get_all_columns(self) -> List[Column]:
        url = 'https://api.github.com/projects/{}/columns?access_token={}'.format(self.project_id, self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)
        if r.status_code == 200:
            return [Column(self.project_id, col) for col in r.json() if col['name'] not in self.cols_ignore]
        return []

    def get_all_cards(self, col: Column) -> List[Card]:
        url = 'https://api.github.com/projects/columns/{}/cards?access_token={}&per_page=100'.format(col.id, self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)
        if r.status_code == 200:
            return [Card(col, card) for card in r.json()]
        return []

    def get_all_current_labels(self, issue) -> List[str]:
        url = 'https://api.github.com/repos/{}/{}/issues/{}/labels?access_token={}'.format(
            issue.owner,
            issue.repo,
            issue.number,
            self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            return [lab['name'] for lab in r.json() if lab['name'] in self.import_label_types]
        return []

    def get_all_events(self, issue_num) -> List[Event]:
        events = []
        for repo in self.repos:
            interesting_event_types = ['labeled', 'unlabeled', 'reopened', 'closed', 'assigned', 'cross-referenced']
            url = 'https://api.github.com/repos/{}/{}/issues/{}/timeline?access_token={}&per_page=100'.format(repo.owner, repo.name, issue_num, self.access_token)
            header = {'Accept': 'application/vnd.github.mockingbird-preview'}
            r = requests_retry_session().get(url, headers=header)

            if r.status_code == 200:

                for ev in r.json():
                    if ev['event'] in interesting_event_types:
                        events.append(Event(ev, self.label2val))
                return events
        return []

    def get_all_active_repos(self, owner, weeks=1) -> List[Repo]:
        rlist = []
        for repo in self.repos:
            url = 'https://api.github.com/orgs/{}/repos?access_token={}'.format(
                repo.owner,
                self.access_token)
            r = requests_retry_session().get(url, headers=self.headers)

            if r.status_code == 200:
                for repo in r.json():
                    last_push = str2date(repo['pushed_at'])
                    if datetime.now()-timedelta(weeks=weeks) < last_push:
                        rlist.append(Repo(repo['name'], repo['id']))
        return rlist

    def get_all_issues(self) -> List[Issue]:
        rlist = []
        for repo in self.repos:
            url = 'https://api.github.com/repos/{}/{}/issues?access_token={}&per_page=100'.format(
                repo.owner,
                repo.name,
                self.access_token)
            r = requests_retry_session().get(url, headers=self.headers)

            if r.status_code == 200:
                for issue_dict in r.json():
                    rlist.append(Issue(repo, issue_dict))
        return rlist

    def get_pr(self, repo, pr_num):
        # /repos/:owner/:repo/pulls/:number
        url = 'https://api.github.com/repos/{}/{}/pulls/{}?access_token={}'.format(
            repo.owner,
            repo.name,
            pr_num,
            self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            pr = PR(repo, r.json())
            revs = self.fetch_PR_reviews(pr)
            pr.add_reviews(revs)
            data = self.fetch_PR_data(pr)
            pr.update(data)
            return pr
        raise ValueError('invalid PR request for {}/{} #{}'.format(repo.owner, repo.name, pr_num))

    def get_all_PRs(self) -> List[PR]:
        pull_requests = []
        for repo in self.repos:
            url = 'https://api.github.com/repos/{}/{}/pulls?access_token={}&per_page=100'.format(
                repo.owner,
                repo.name,
                self.access_token)
            params = {
                'state': 'all',
                'sort': 'updated',
                'direction': 'desc'
            }
            r = requests_retry_session().get(url, headers=self.headers, data=params)

            if r.status_code == 200:
                for pr_dict in r.json():
                    pull_requests.append(PR(repo, pr_dict))
        return pull_requests

    def fetch_PR_data(self, pr: PR) -> List[Review]:
        # GET /repos/:owner/:repo/pulls/:number/reviews
        url = 'https://api.github.com/repos/{}/{}/pulls/{}?access_token={}'.format(
            pr.repo.owner,
            pr.repo.name,
            pr.number,
            self.access_token)

        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            return r.json()
        return None

    def fetch_PR_reviews(self, pr: PR) -> List[Review]:
        # GET /repos/:owner/:repo/pulls/:number/reviews
        url = 'https://api.github.com/repos/{}/{}/pulls/{}/reviews?access_token={}&per_page=100'.format(
            pr.repo.owner,
            pr.repo.name,
            pr.number,
            self.access_token)

        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            revs = []
            for rev in r.json():
                revs.append(Review(pr, rev))
            return revs
        return []

    def get_single_issue_from_card(self, card: Card) -> Issue:
        # /repos/:owner/:repo/issues/:number
        url = 'https://api.github.com/repos/{}/{}/issues/{}?access_token={}'.format(
            card.repo.owner,
            card.repo.name,
            card.issue_num,
            self.access_token)

        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            return Issue(card.repo, r.json())
        return None

    def get_production_sha(self, repo: Repo) -> str:
        # GET /repos/:owner/:repo/git/refs/tags
        ref = 'production'
        url = 'https://api.github.com/repos/{}/{}/git/refs/tags?access_token={}'.format(
            repo.owner,
            repo.name,
            self.access_token)

        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            prod_tag = [r for r in r.json() if 'refs/tags/{}'.format(ref)
                        == r['ref']][0]
            commit_sha = prod_tag['object']['sha']
            return commit_sha
        return None

    def get_master_sha(self, repo: Repo) -> str:
        # GET #/repos/:owner/:repo/git/refs/:ref
        url = 'https://api.github.com/repos/{}/{}/git/refs?access_token={}'.format(
            repo.owner,
            repo.name,
            self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)
        if r.status_code == 200:
            master_ref = [r for r in r.json() if 'master' in r['ref']][0]
            commit_sha = master_ref['object']['sha']
            return commit_sha
        return None

    def compare_commits(self, repo: Repo, base: str, head: str) -> List[Commit]:
        # /repos/:owner/:repo/compare/:base...:head
        url = 'https://api.github.com/repos/{}/{}/compare/{}...{}?access_token={}'.format(
            repo.owner,
            repo.name,
            base,
            head,
            self.access_token)
        r = requests_retry_session().get(url, headers=self.headers)

        if r.status_code == 200:
            cmp = r.json()

            return cmp['ahead_by'], [Commit(c) for c in cmp['commits']]
        return None, None



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ghsprint',
    packages=['ghsprint'],
    version='0.5',
    license='GNU GPLv3',
    description='Contains utilities for agile sprints that use the GitHub platform.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Alexander Jahn',
    author_email='jahn.alexander@gmail.com',
    url='https://github.com/AlxndrJhn/ghsprint',
    download_url='https://github.com/AlxndrJhn/ghsprint/archive/0.5.tar.gz',
    keywords=['GitHub', 'utilities', 'reporting', 'sprint'],
    install_requires=[
        'requests',
        'urllib3',
        'rake-nltk',
        'tqdm',
        'click',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Utilities',

        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

    ],
)



from datetime import datetime
from typing import List, Dict

from .board_stuff import Column
from .issue_event import Event
from .utils import str2date
from .repo_stuff import Repo


class Card(object):
    def __init__(self, col: Column, input_dict: dict):
        self.col = col
        self.id = input_dict['id']
        self.created_at = str2date(input_dict['created_at'])
        self.updated_at = str2date(input_dict['updated_at'])
        self.content_url = input_dict.get('content_url', None)

        self.has_issue = False
        self.issue_num = None
        self.repo = None
        if self.content_url:
            self.repo = Repo.from_http(self.content_url)
            self.has_issue = True
            self.issue_num = int(self.content_url.split('/')[-1])
        self.events = []

    def __str__(self):
        return 'id: {}, issue_num: {}'.format(self.id, self.issue_num)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.created_at == other.created_at and self.content_url == other.content_url

    def set_events(self, events: List[Event]):
        self.events = events

    def get_state(self):
        last_change = next((ev.event for ev in self.events[::-1] if ev.event in ['reopened', 'closed']), None)
        state = 'open'
        if last_change:
            state = last_change

        return state

    def has_label_val_created_within(self, start: datetime, end: datetime, min_value: float = 0.5):
        val = self.get_pokered_value(start, end)
        return val is not None and val >= min_value

    def get_pokered_value(self, start: datetime, end: datetime):
        vals = [e.label_val for e in self.events if start <=
                e.date <= end and e.label_val]
        if len(vals) > 0:
            return vals[0]
        else:
            return None

    def get_current_value(self):
        vals = [e.label_val for e in self.events if e.label_val]
        if len(vals) > 0:
            return vals[-1]
        else:
            return None

    def set_issue(self, issue):
        self.issue = issue

    def was_assigned(self):
        return any(o.assignee for o in self.events)

    def get_assignees(self, login_to_name_mapping: Dict[str, str]={}):
        assignees = []
        for user in self.issue.assignees:
            user_name = login_to_name_mapping.get(user['login'].lower(), user['login'])
            assignees.append(user_name)

        return assignees or ['not assigned']



from enum import Enum

from typing import Dict

from .issue import Issue, states
from .repo_stuff import Repo
from .utils import str2date


class PR(object):
    def __init__(self, repo: Repo, pr_dict: dict):
        self.repo = repo
        self.id = pr_dict['id']
        self.number = pr_dict['number']
        self.title = pr_dict['title']
        self.state = states[pr_dict['state']]
        self.created_at = str2date(pr_dict['created_at'])
        self.updated_at = str2date(pr_dict['updated_at'])
        self.url = pr_dict['html_url']

        self.merged_at = str2date(pr_dict['merged_at'])
        self.is_in_prod = None
        self.body = pr_dict['body']
        self.merge_commit_sha = pr_dict['merge_commit_sha']
        self.user_name = pr_dict['user']['login']

        self.labels = []
        self.deletions = None
        self.additions = None
        self.changed_files = None
        self.is_draft = False

        self.closes = []
        closes_lines = [line.strip() for line in pr_dict['body'].lower().split('\r\n') if line.startswith('- closes ')]
        if len(closes_lines) > 0:
            for line in closes_lines:
                issue = Issue.from_closes_line(line, repo)
                if issue:
                    self.closes.append(issue)

    def add_reviews(self, revs):
        self.reviews = revs

    def update(self, data):
        self.deletions = data['deletions']
        self.additions = data['additions']
        self.changed_files = data['changed_files']
        self.labels = data['labels']
        self.is_draft = data['mergeable_state'] == 'draft'

    def get_state(self):
        state = str(self.state.name)
        if self.merged_at:
            state = 'merged'
        return state

    def get_review_state(self, login_to_name_mapping: Dict[str, str]={}):
        ignore_states = [Review.states.DISMISSED, Review.states.COMMENTED]
        unique_reviewers = list({o.reviewer for o in self.reviews if o.state not in ignore_states})
        if len(unique_reviewers) == 0:
            return 'not reviewed'
        txt = []
        for reviewer in unique_reviewers:
            latest_review = next(rev for rev in self.reviews[::-1] if rev.reviewer == reviewer and rev.state not in ignore_states)
            mapped_reviewer_name = login_to_name_mapping.get(reviewer.lower(), reviewer)
            txt.append(f'{mapped_reviewer_name} > {latest_review.state.name.lower().replace("_", " ")}')
        return ', '.join(txt)


class Review(object):
    class states(Enum):
        CHANGES_REQUESTED = 0
        COMMENTED = 1
        APPROVED = 2
        DISMISSED = 3
        PENDING = 4

    def __init__(self, pr: PR, rev: dict):
        self.pr = pr
        self.submitted_at = str2date(rev.get('submitted_at', None))
        self.url = rev['html_url']
        self.reviewer = rev['user']['login']
        self.state = Review.states[rev['state']]
